% % =======================================================================
% % This code computes the rare-event probability of large portfolio loss
% % in the t copula model under a variety of settings
% %
% % See: 
% % Chan, J. C. C. and Kroese, D. K. (2012). "Improved Cross-Entropy Method
% % for Estimation", Statistics and Computing, 22(5), 1031-1040.
% %
% % Please report any errors to joshuacc.chan@gmail.com
% % =======================================================================

clear; clc;
method = 1;               % 1: CE; 2: VM
m = 5;                    % # of factors
n = 250;                  % # of obligators
b = .25;
nu = 12;                  % df for the t-dist
x = .5 * sqrt(n);         % individual threshold
l = n*b;                  % overall threshold
a = 1;                    % exposure of lender
rho = .25;                % weight of common risk factor
sig_eta = 9;

nloop = 1000;             % # of MCMC iterations
burnin = 50;
N = 50000;                % Monte Carlo sample size

sq1mrho2 = sqrt(1-rho^2);
sZ = zeros(nloop-burnin,m); 
sQ = zeros(nloop-burnin,m);
sL = zeros(nloop-burnin,m);
seta = zeros(nloop-burnin,m,n);
counteta = 0;

start_time = clock;
%% initialize
Z = randn(1,m);
eta = sqrt(sig_eta) * randn(n,m);
sortH = sort((repmat(rho*Z,n,1) + sqrt(1-rho^2)*eta)/x);
H = max(sortH(n-floor(1/a*l),:).^2,0);
Q = zeros(1,m);
for i=1:m
    Q(i) =  rtgamrnd(nu/2,1/2,H(i)*nu,1)/nu;
end
sqQ = sqrt(Q);
X = (repmat(rho*Z,n,1) + sqrt(1-rho^2)*eta)./repmat(sqQ,n,1);
L = sum(a * (X>x));

for loop=1:nloop

    % sample Z|L(X)>gam, eta, Q
    sortH = sort((repmat(x*sqQ,n,1) - sq1mrho2*eta)/rho);
    H = sortH(floor(1/a*l)+1,:);
    for i=1:m
        Z(i) = ltnormrnd(0,1,H(i),1);
    end
    
    % sample Q|L(X)>gam, eta, Z
    sortH = sort((repmat(rho*Z,n,1) + sqrt(1-rho^2)*eta)/x);
    H = max(sortH(n-floor(1/a*l),:).^2,0);
    Q = zeros(1,m);
    for i=1:m
        Q(i) =  rtgamrnd(nu/2,1/2,H(i)*nu,1)/nu;
    end
    sqQ = sqrt(Q);
    
    % sample eta|L(X)>gam, Z, Q        
    etac = sqrt(sig_eta) * randn(n,m);
    X = (repmat(rho*Z,n,1) + sq1mrho2*etac)./repmat(sqQ,n,1);
    L = sum(a*(X>x));
    for i=1:m        
        if L(i)>l
            eta(:,i) = etac(:,i);
        end
    end

    if loop>burnin
        sZ(loop-burnin,:) = Z;
        sQ(loop-burnin,:) = Q;
        X = (repmat(rho*Z,n,1) + sq1mrho2*eta)./repmat(sqQ,n,1);
        L = sum(a*(X>x));
        sL(loop-burnin,:) = L;
        seta(loop-burnin,:,:) = eta';   
    end    
end

disp( ['MCMC takes '  num2str( etime(clock, start_time) ) ' seconds' ] );

sZ = reshape(sZ,m*(nloop-burnin),1);
sQ = reshape(sQ,m*(nloop-burnin),1);
seta = reshape(seta,m*(nloop-burnin),n);
sL = reshape(sL,m*(nloop-burnin),1);

if method == 1 % use CE
    tmodel_IS2  
elseif method == 2  % use VM
    tmodel_VM  
end

fprintf('\n'); 
fprintf('Estimate | Relative error:\n'); 
fprintf('%.8f| %.1f\n',lhat,RE);
